import numpy as np

from scipy.optimize import minimize
from pypop7.optimizers.pso.pso import PSO


class IPSOLS(PSO):
    """Incremental particle swarm optimizer (IPSO).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - swarm (population) size, number of particles (`int`, default: `20`),
                * 'cognition'     - cognitive learning rate (`float`, default: `2.0`),
                * 'society'       - social learning rate (`float`, default: `2.0`),
                * 'max_ratio_v'   - maximal ratio of velocities w.r.t. search range (`float`, default: `0.2`).

    Attributes
    ----------
    n_individuals : `int`
                    swarm (population) size, number of particles.
    cognition     : `float`
                    cognitive learning rate.
    society       : `float`
                    social learning rate.
    max_ratio_v   : `float`
                    maximal ratio of velocities w.r.t. search range.

    References
    ----------
    De Oca, M.A.M., Stutzle, T., Van den Enden, K. and Dorigo, M., 2010.
    Incremental social learning in particle swarms.
    IEEE Transactions on Systems, Man, and Cybernetics, Part B (Cybernetics), 41(2), pp.368-384.
    https://ieeexplore.ieee.org/document/5582312
    """

    def __init__(self, problem, options):
        PSO.__init__(self, problem, options)
        self.n_individuals = 1
        # max swarm (population) size, number of particles
        self.max_n_individuals = options.get('max_n_individuals', 1000)
        self.cognition = options.get('cognition', 2.05)  # cognitive learning rate
        self.society = options.get('society', 2.05)  # social learning rate
        self.constriction = options.get('constriction', 0.729)  # constriction factor
        # for local search
        self.e = np.ones((self.n_individuals,))  # Whether a local search should be invoked for particles or not
        self.powell_tolerance = 0.01  # Powell's method tolerance
        self.powell_max_iterations = 10  # Powell's method maximum number of iterations
        self.powell_step_size = 0.2  # Powell's method step size: 20% of the length of the search range

    def initialize(self, args=None):
        v = np.zeros((self.n_individuals, self.ndim_problem))  # velocities
        x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                            size=self._swarm_shape)  # positions
        y = np.empty((self.n_individuals,))  # fitness
        p_x, p_y = np.copy(x), np.copy(y)  # personally previous-best positions and fitness
        n_x = np.copy(x)  # neighborly previous-best positions
        for i in range(self.n_individuals):
            if self._check_terminations():
                return v, x, y, p_x, p_y, n_x
            res = minimize(self.fitness_function, x[i], method="Powell",
                           tol=self.powell_tolerance, options={"maxiter": self.powell_max_iterations})
            self.n_function_evaluations += (res.nfev - 1)
            x[i] = res.x
            y[i] = self._evaluate_fitness(x[i], args)
        p_y = np.copy(y)
        return v, x, y, p_x, p_y, n_x

    def iterate(self, v=None, x=None, y=None, p_x=None, p_y=None, n_x=None, args=None):
        for i in range(self.n_individuals):
            if self.e[i]:
                if self._check_terminations():
                    return v, x, y, p_x, p_y, n_x
                res = minimize(self.fitness_function, x[i], method="Powell",
                               tol=self.powell_tolerance, options={"maxiter": self.powell_max_iterations})
                self.n_function_evaluations += (res.nfev - 1)
                x[i] = res.x
                y[i] = self._evaluate_fitness(x[i], args)
                if y[i] < p_y[i]:
                    p_x[i], p_y[i] = x[i], y[i]
                if res.success:
                    self.e[i] = 0

        # Horizontal social learning
        for i in range(self.n_individuals):
            if self._check_terminations():
                return v, x, y, p_x, p_y, n_x
            n_x[i] = p_x[np.argmin(p_y)]  # online update within global topology
            cognition_rand = self.rng_optimization.uniform(size=(self.ndim_problem,))
            society_rand = self.rng_optimization.uniform(size=(self.ndim_problem,))
            v[i] = self.constriction * (v[i] + self.cognition * cognition_rand * (p_x[i] - x[i]) +
                                        self.society * society_rand * (n_x[i] - x[i]))  # velocity update
            min_v, max_v = v[i] < self._min_v, v[i] > self._max_v
            v[i, min_v], v[i, max_v] = self._min_v[min_v], self._max_v[max_v]
            x[i] += v[i]  # position update
            y[i] = self._evaluate_fitness(x[i], args)  # fitness evaluation
            if y[i] < p_y[i]:  # online update
                p_x[i], p_y[i] = x[i], y[i]
                self.e[i] = 1  # not converged to a local optimum

        # Population growth and vertical social learning
        if self.n_individuals < self.max_n_individuals:
            if self._check_terminations():
                return v, x, y, p_x, p_y, n_x
            xx = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary)
            xx = xx + self.rng_initialization.uniform(size=(self.ndim_problem,)) * (n_x[-1] - xx)
            yy = self._evaluate_fitness(xx, args)  # fitness evaluation
            v = np.vstack((v, np.zeros((self.ndim_problem,))))
            x, y = np.vstack((x, xx)), np.hstack((y, yy))
            p_x, p_y = np.vstack((p_x, xx)), np.hstack((p_y, yy))
            n_x = np.vstack((n_x, p_x[np.argmin(p_y)]))
            self.e = np.hstack((self.e, 1))
            self.n_individuals += 1
        self._n_generations += 1
        return v, x, y, p_x, p_y, n_x



"""Repeat the following paper for `ESA`:
    Siarry, P., Berthiau, G., Durdin, F. and Haussy, J., 1997.
    Enhanced simulated annealing for globally minimizing functions of many-continuous variables.
    ACM Transactions on Mathematical Software, 23(2), pp.209-228.
    https://dl.acm.org/doi/abs/10.1145/264029.264043

    Since several details were not very clear in the original paper, it is hard to
      repeat the original experiments *perfectly*, according to our experiences.
    Our current implementation could generate the *relatively close but not perfect* result,
      depending on your level of acceptance.
    We still expect to improve it to match the original paper in the future (if possible).

    The fuzzy details are presented below, which can cause different implementations with different performances:
    1. Initial FOBJ variation average, DGYINI, obtained from (typically) 50 uphill moves performed
      before beginning SA minimization itself.
    2. Step 2: Space Partitioning.
    3. Those elementary operations on only p coordinates move the point XSTART to the point XTRY
      satisfying all hyperrectangular box constraints.
    4. MSOTST = 0 (this is a typo error in the original paper).
    5. Other systematic experiments are still in progress in order to obtain, if possible, a rule for
      the p-optimal choice, adapted to FOBJ-specific characteristics.
"""
import numpy as np

from pypop7.benchmarks.base_functions import rosenbrock
from pypop7.optimizers.sa.esa import ESA


if __name__ == '__main__':
    problem = {'fitness_function': rosenbrock,
               'ndim_problem': 50,
               'upper_boundary': 10 * np.ones((50,)),
               'lower_boundary': -5 * np.ones((50,))}
    options = {'max_function_evaluations': 78224,
               'seed_rng': 1,  # undefined in the original paper
               'p': 10,  # tuned
               'saving_fitness': 1000,
               'verbose': 10000}
    esa = ESA(problem, options)
    results = esa.optimize()
    # FOBJ initial-value average: 6472239 (from 50 samples) vs 5e5 (from the original paper)
    # our implementation starts from a worse starting point than the original paper.
    print(results)  # 1.31007965e+01 vs 8.8 (from the original paper)



"""Repeat the following paper for `CCPSO2`:
    Li, X. and Yao, X., 2012.
    Cooperatively coevolving particle swarms for large scale optimization.
    IEEE Transactions on Evolutionary Computation, 16(2), pp.210-224.
    https://ieeexplore.ieee.org/document/5910380/

    Luckily our code could repeat the data generated by the original paper *well*.
    Therefore, we argue that the repeatability of `CCPSO2` could be **well-documented**.
"""
import time

import numpy as np

from pypop7.benchmarks.base_functions import sphere, rastrigin, griewank
from pypop7.optimizers.pso.ccpso2 import CCPSO2 as Solver


if __name__ == '__main__':
    ndim_problem = 1000

    start_run = time.time()
    problem = {'fitness_function': sphere,
               'ndim_problem': ndim_problem,
               'lower_boundary': -100 * np.ones((ndim_problem,)),
               'upper_boundary': 100 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 5e6,
               'seed_rng': 0,
               'verbose': 1e1}
    solver = Solver(problem, options)
    results = solver.optimize()
    print(results)  # 1.472676192640866e-28 vs 5.18e-13 (from the original paper)
    print('*** Runtime: {:7.5e}'.format(time.time() - start_run))

    start_run = time.time()
    problem = {'fitness_function': rastrigin,
               'ndim_problem': ndim_problem,
               'lower_boundary': -5 * np.ones((ndim_problem,)),
               'upper_boundary': 5 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 5e6,
               'seed_rng': 0,
               'verbose': 1e1}
    solver = Solver(problem, options)
    results = solver.optimize()
    print(results)  # 0.2947450894425856 vs 1.99e-01 (from the original paper)
    print('*** Runtime: {:7.5e}'.format(time.time() - start_run))

    start_run = time.time()
    problem = {'fitness_function': griewank,
               'ndim_problem': ndim_problem,
               'lower_boundary': -600 * np.ones((ndim_problem,)),
               'upper_boundary': 600 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 5e6,
               'seed_rng': 0,
               'verbose': 1e1}
    solver = Solver(problem, options)
    results = solver.optimize()
    print(results)  # 4.218847493575595e-15 vs 1.18e-03 (from the original paper)
    print('*** Runtime: {:7.5e}'.format(time.time() - start_run))



"""Repeat the following paper for `PRS`:
    Hansen, N., Arnold, D.V. and Auger, A., 2015.
    Evolution strategies.
    In Springer Handbook of Computational Intelligence (pp. 871-898).
    Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007%2F978-3-662-43505-2_44

    All generated figures can be accessed via the following link:
    https://github.com/Evolutionary-Intelligence/pypop/blob/main/docs/repeatability/prs/_repeat_prs.png

    Luckily our code could repeat the data reported in the original paper *well*.
    Therefore, we argue that the repeatability of `PRS` could be **well-documented**.
"""
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.base_functions import sphere
from pypop7.optimizers.rs.prs import PRS


if __name__ == '__main__':
    sns.set_theme(style='darkgrid')
    plt.figure()
    ndim = 10
    for i in range(3):
        problem = {'fitness_function': sphere,
                   'ndim_problem': ndim,
                   'upper_boundary': 1 * np.ones((ndim,)),
                   'lower_boundary': -0.2 * np.ones((ndim,))}
        options = {'max_function_evaluations': 1500,
                   'seed_rng': i,  # undefined in the original paper
                   'saving_fitness': 1}
        prs = PRS(problem, options)
        results = prs.optimize()
        fitness = results['fitness']
        plt.plot(fitness[:, 0], fitness[:, 1], 'r')
    plt.xticks([0, 500, 1000, 1500])
    plt.xlim([0, 1500])
    plt.yscale('log')
    plt.yticks([1e-9, 1e-6, 1e-3, 1e0])
    plt.show()



"""Repeat the following paper for `TDE`:
    Fan, H.Y. and Lampinen, J., 2003.
    A trigonometric mutation operation to differential evolution.
    Journal of Global Optimization, 27(1), pp.105-129.
    https://link.springer.com/article/10.1023/A:1024653025686

    All generated figures can be accessed via the following link:
    https://github.com/Evolutionary-Intelligence/pypop/tree/main/docs/repeatability/tde

    Luckily our code could repeat the data reported in the original paper *well*.
    Therefore, we argue that the repeatability of `TDE` could be **well-documented**.
"""
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.base_functions import ackley, rastrigin
from pypop7.optimizers.de.tde import TDE as Solver


if __name__ == '__main__':
    sns.set_theme(style='darkgrid')
    for f in [ackley, rastrigin]:
        plt.figure()
        print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
        problem, options = {}, {}
        if f.__name__ == 'ackley':
            ndim_problem = 30
            problem = {'fitness_function': f,
                       'ndim_problem': ndim_problem,
                       'lower_boundary': -20 * np.ones((ndim_problem,)),
                       'upper_boundary': 30 * np.ones((ndim_problem,))}
            options = {'max_runtime': 45,
                       'seed_rng': 0,  # undefined in the original paper
                       'saving_fitness': 1,
                       'verbose': 2000}
        elif f.__name__ == 'rastrigin':
            ndim_problem = 20
            problem = {'fitness_function': f,
                       'ndim_problem': ndim_problem,
                       'lower_boundary': -5.12 * np.ones((ndim_problem,)),
                       'upper_boundary': 5.12 * np.ones((ndim_problem,))}
            options = {'max_runtime': 30,
                       'seed_rng': 0,  # undefined in the original paper
                       'saving_fitness': 1,
                       'verbose': 2000}
        solver = Solver(problem, options)
        results = solver.optimize()
        results['fitness'][:, 0] *= results['runtime']/results['n_function_evaluations']
        if f.__name__ == 'ackley':
            plt.ylim([0, 20])
            plt.yticks([0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20])
            plt.xlim([0, 45])
            plt.xticks([0, 5, 10, 15, 20, 25, 30, 35, 40, 45])
            plt.ylabel("f1(X)")
        elif f.__name__ == 'rastrigin':
            plt.ylim([0, 160])
            plt.yticks([0, 20, 40, 60, 80, 100, 120, 140, 160])
            plt.xlim([0, 30])
            plt.xticks([0, 5, 10, 15, 20, 25, 30])
            plt.ylabel("f2(x)")
        plt.plot(results['fitness'][:, 0], results['fitness'][:, 1], color='black')
        plt.xlabel("CPU Time (Seconds)")
        plt.title(f.__name__.capitalize())
        plt.show()



import unittest

from pypop7.benchmarks.base_functions import *
from test_cases import *


class TestBaseFunctions(unittest.TestCase):
    def test_squeeze_and_check(self):
        self.assertEqual(squeeze_and_check(0), np.array([0]))
        self.assertEqual(squeeze_and_check(np.array(0)), np.array([0]))
        x1 = np.array([0.7])
        self.assertEqual(squeeze_and_check(x1), x1)
        x2 = np.array([0.0, 1.0])
        self.assertTrue(np.allclose(squeeze_and_check(x2), x2))
        x3 = np.arange(6).reshape(2, 3)
        with self.assertRaisesRegex(TypeError, 'The number+'):
            squeeze_and_check(x3)
        with self.assertRaisesRegex(TypeError, 'The size should > 1+'):
            squeeze_and_check(x1, True)
        with self.assertRaisesRegex(TypeError, 'the size should != 0.'):
            squeeze_and_check([])

    def test_sphere(self):
        sample = TestCases()
        for func in [sphere, Sphere()]:
            for ndim in range(1, 8):
                self.assertTrue(sample.compare(func, ndim, get_y_sphere(ndim - 1)))
            self.assertTrue(sample.check_origin(func))

    def test_cigar(self):
        sample = TestCases()
        for func in [cigar, Cigar()]:
            for ndim in range(2, 8):
                self.assertTrue(sample.compare(func, ndim, get_y_cigar(ndim - 2)))
            with self.assertRaisesRegex(TypeError, 'The size should > 1+'):
                sample.compare(func, 1, np.empty((5,)))
            self.assertTrue(sample.check_origin(func))

    def test_discus(self):
        sample = TestCases()
        for func in [discus, Discus()]:
            for ndim in range(2, 8):
                self.assertTrue(sample.compare(func, ndim, get_y_discus(ndim - 2)))
            with self.assertRaisesRegex(TypeError, 'The size should > 1+'):
                sample.compare(func, 1, np.empty((5,)))
            self.assertTrue(sample.check_origin(func))

    def test_cigar_discus(self):
        sample = TestCases()
        for func in [cigar_discus, CigarDiscus()]:
            for ndim in range(2, 8):
                self.assertTrue(sample.compare(func, ndim, get_y_cigar_discus(ndim - 2)))
            with self.assertRaisesRegex(TypeError, 'The size should > 1+'):
                sample.compare(func, 1, np.empty((5,)))
            self.assertTrue(sample.check_origin(func))

    def test_ellipsoid(self):
        sample = TestCases()
        for func in [ellipsoid, Ellipsoid()]:
            for ndim in range(2, 8):
                self.assertTrue(sample.compare(func, ndim, get_y_ellipsoid(ndim - 2)))
            with self.assertRaisesRegex(TypeError, 'The size should > 1+'):
                sample.compare(func, 1, np.empty((5,)))
            self.assertTrue(sample.check_origin(func))

    def test_different_powers(self):
        sample = TestCases()
        for func in [different_powers, DifferentPowers()]:
            for ndim in range(2, 8):
                self.assertTrue(sample.compare(func, ndim, get_y_different_powers(ndim - 2), atol=0.1))
            with self.assertRaisesRegex(TypeError, 'The size should > 1+'):
                sample.compare(func, 1, np.empty((5,)))
            self.assertTrue(sample.check_origin(func))

    def test_schwefel221(self):
        sample = TestCases()
        for func in [schwefel221, Schwefel221()]:
            for ndim in range(1, 8):
                self.assertTrue(sample.compare(func, ndim, get_y_schwefel221(ndim - 1)))
            self.assertTrue(sample.check_origin(func))

    def test_step(self):
        sample = TestCases()
        for func in [step, Step()]:
            for ndim in range(1, 8):
                self.assertTrue(sample.compare(func, ndim, get_y_step(ndim - 1)))
            self.assertTrue(sample.check_origin(func))

    def test_schwefel222(self):
        sample = TestCases()
        for func in [schwefel222, Schwefel222()]:
            for ndim in range(1, 8):
                self.assertTrue(sample.compare(func, ndim, get_y_schwefel222(ndim - 1)))
            self.assertTrue(sample.check_origin(func))

    def test_rosenbrock(self):
        sample = TestCases()
        for func in [rosenbrock, Rosenbrock()]:
            for ndim in range(2, 8):
                self.assertTrue(sample.compare(func, ndim, get_y_rosenbrock(ndim - 2)))
            with self.assertRaisesRegex(TypeError, 'The size should > 1+'):
                sample.compare(func, 1, np.empty((5,)))

    def test_schwefel12(self):
        sample = TestCases()
        for func in [schwefel12, Schwefel12()]:
            for ndim in range(2, 8):
                self.assertTrue(sample.compare(func, ndim, get_y_schwefel12(ndim - 2)))
            with self.assertRaisesRegex(TypeError, 'The size should > 1+'):
                sample.compare(func, 1, np.empty((5,)))
            self.assertTrue(sample.check_origin(func))

    def test_exponential(self):
        for func in [exponential, Exponential()]:
            for ndim in range(1, 8):
                self.assertTrue(np.abs(func(np.zeros((ndim,))) + 1) < 1e-9)

    def test_griewank(self):
        sample = TestCases()
        for func in [griewank, Griewank()]:
            for ndim in range(2, 8):
                self.assertTrue(sample.compare(func, ndim, get_y_griewank(ndim - 2), atol=0.001))
            self.assertTrue(sample.check_origin(func))

    def test_bohachevsky(self):
        sample = TestCases()
        for func in [bohachevsky, Bohachevsky()]:
            for ndim in range(1, 5):
                self.assertTrue(sample.compare(func, ndim, get_y_bohachevsky(ndim - 1), atol=0.1))
            self.assertTrue(sample.check_origin(func))

    def test_ackley(self):
        sample = TestCases()
        for func in [ackley, Ackley()]:
            for ndim in range(2, 8):
                self.assertTrue(sample.compare(func, ndim, get_y_ackley(ndim - 2), atol=0.001))
            self.assertTrue(sample.check_origin(func))

    def test_rastrigin(self):
        sample = TestCases()
        for func in [rastrigin, Rastrigin()]:
            for ndim in range(2, 8):
                self.assertTrue(sample.compare(func, ndim, get_y_rastrigin(ndim - 2)))
            self.assertTrue(sample.check_origin(func))

    def test_levy_montalvo(self):
        for func in [levy_montalvo, LevyMontalvo()]:
            for ndim in range(1, 8):
                self.assertTrue(np.abs(func(-np.ones((ndim,)))) < 1e-9)

    def test_michalewicz(self):
        sample = TestCases()
        for func in [michalewicz, Michalewicz()]:
            self.assertTrue(sample.check_origin(func))

    def test_salomon(self):
        sample = TestCases()
        for func in [salomon, Salomon()]:
            self.assertTrue(sample.check_origin(func))

    def test_shubert(self):
        minimizers = [[-7.0835, 4.858], [-7.0835, -7.7083], [-1.4251, -7.0835], [5.4828, 4.858],
                      [-1.4251, -0.8003], [4.858, 5.4828], [-7.7083, -7.0835], [-7.0835, -1.4251],
                      [-7.7083, -0.8003], [-7.7083, 5.4828], [-0.8003, -7.7083], [-0.8003, -1.4251],
                      [-0.8003, 4.8580], [-1.4251, 5.4828], [5.4828, -7.7083], [4.858, -7.0835],
                      [5.4828, -1.4251], [4.858, -0.8003]]
        for func in [shubert, Shubert()]:
            for minimizer in minimizers:
                self.assertTrue((np.abs(func(minimizer) + 186.7309) < 1e-3))


if __name__ == '__main__':
    unittest.main()
